/**
 * GTK-Programm, welches mit dem per Cairo gezeichneten Lochstreifen
 * Dateien visualisieren kann.
 *
 * Dabei wird die Flexibilitaet des Cairo-Zeichners voll ausgeschoepft:
 * Beliebige Dateien koennen zum Anzeigen geoeffnet werden, der Loch-
 * streifen kann in saemtlichen Farben frei variiert werden, die
 * anzuzeigenden Emptybytes koennen beliebig gesetzt werden.
 * Die Darstellung kann jederzeit in jede Richtung gedreht und gespiegelt
 * werden, ausserdem ist stufenloser Zoom sowie eine automatische
 * Groessenanpassung moeglich, sodass der Lochstreifen sich immer genau an die
 * Fenstergroesse anpasst und man sinnvoll einfach ueber die Daten "scrollen"
 * kann.
 *
 * Die aktuelle Ansicht des Lochstreifens kann einfach als PNG-Bild oder
 * SVG-Vektorgrafik in eine Datei exportiert werden. Die Aenderung der Datei
 *  la "Hexeditor" ist nicht vorgesehen, dafuer werden aber beim Ueberfahren
 * des Lochstreifens in der Statusleiste die Werte des Bytes, ueber dem man
 * sich befindet, aufgeschluesselt.
 *
 * Das Programm ist bewusst unaufdringlich gehalten; die meisten Nachrichten
 * werden ueber die Statuszeile ausgegeben. Die Bedienung erfolgt massgeblich
 * ueber das Menue.
 *
 * Je nach Kommandozeilenparameter wird von STDIN gelesen oder eine Datei
 * geoeffnet (mehr Informationen mit Aufruf --help).
 * Wird gar kein Parameter angegeben, dann wird eine kleine Startanimation
 * angezeigt, die von der Perfomance und Flexibilitaet des Zeichenprogramms
 * zeugt ;-)
 * 
 * -- Sven, November 2007
 *
 **/
#include <stdio.h>
#include <stdlib.h>
#include <math.h> // rint()
#include <string.h> // strlen()
#include <gtk/gtk.h>
#include "gtkpapertape.h"

#include <time.h> // NUR WAEHREND DEBUGGING 

#define WINDOW_TITLE "Lochstreifen-Visualisierung" /* Standardtitel */

// Muss leider global sein, damit ein Zugriff von mehreren Funktionen moeglich ist
GtkWidget *window; // Hauptfenster
GTimer *last_statusbar_update; // fuer Statusbar-Prioritaetenbehandlung
GtkWidget *fit_screen_toggle; // Auswahlbox im Ansichtmenue, um Autozoom umzuschalten.
gboolean startsequence_running = FALSE; // Ist TRUE, wenn die Startanimation laeuft

// Die Daten, die bei der Startsequenz angezeigt werden, liegen hier als Bytearray
// vor. Generierung mit
// $ [generatorprogramm] | hexdump -e '11/1 "0x%02x, " "\n"'
// "0x  , "-Zeichen wegschneiden.
// Laenge feststellen mit wc -c

int startsequence_length = 205;
byte_t startsequence_data[] = {
0xfc, 0x01, 0x02, 0x01, 0xfc, 0x00, 0x81, 0xff, 0x81, 0x00, 0xff,
0x01, 0x01, 0x00, 0xff, 0x01, 0x01, 0x00, 0xff, 0x18, 0xc7, 0x00,
0xff, 0x81, 0xff, 0x00, 0xff, 0x80, 0x60, 0x80, 0xff, 0x00, 0xff,
0x80, 0x60, 0x80, 0xff, 0x00, 0xff, 0x91, 0x91, 0x00, 0xff, 0x40,
0x3c, 0x02, 0xff, 0x00, 0x00, 0x00, 0x83, 0xb9, 0xc1, 0x00, 0xff,
0x01, 0xff, 0x00, 0xff, 0x80, 0x60, 0x80, 0xff, 0x00, 0x00, 0x00,
0x81, 0xff, 0x81, 0x00, 0xff, 0x40, 0x3c, 0x02, 0xff, 0x00, 0x80,
0xff, 0x80, 0x00, 0xff, 0x91, 0x91, 0x00, 0xff, 0x88, 0xf7, 0x00,
0xff, 0x88, 0x88, 0xff, 0x00, 0xff, 0x18, 0xc7, 0x00, 0x80, 0xff,
0x80, 0x00, 0x81, 0xff, 0x81, 0x00, 0xfc, 0x01, 0xfc, 0x00, 0xff,
0x91, 0x91, 0x00, 0xff, 0x40, 0x3c, 0x02, 0xff, 0x00, 0x00, 0x00,
0xff, 0x01, 0x01, 0x00, 0xff, 0x81, 0xff, 0x00, 0xff, 0x81, 0x81,
0x00, 0xff, 0x10, 0xff, 0x00, 0xf0, 0x91, 0x9f, 0x00, 0x80, 0xff,
0x80, 0x00, 0xff, 0x88, 0xf7, 0x00, 0xff, 0x91, 0x91, 0x00, 0x81,
0xff, 0x81, 0x00, 0xff, 0x90, 0x90, 0x00, 0xff, 0x91, 0x91, 0x00,
0xff, 0x40, 0x3c, 0x02, 0xff, 0x00, 0x83, 0xb9, 0xc1, 0x00, 0xff,
0x91, 0x91, 0x00, 0x81, 0xff, 0x81, 0x00, 0xff, 0x81, 0x81, 0x00,
0xff, 0x10, 0xff, 0x00, 0xff, 0x40, 0x3c, 0x02, 0xff, 0x00, 0xff,
0x91, 0x91, 0x00, 0xff, 0x88, 0xf7, 0x00
};

// zum Testen:
/*int startsequence_length = 4;
byte_t startsequence_data[] = { 0x00, 0x01, 0x02, 0x03 };*/

void message_statusbar(char *msg);
void message_error(gchar *heading, gchar *text);


gboolean open_lochstreifen_file(char *filename) {
    /**
     * Oeffnet die gewuenschte Datei.
     * Bei Fehler wird FALSE zurueckgegeben
     **/
     int length; gchar *data;
     gboolean ret;
     GError *err;

     ret = g_file_get_contents(filename, &data, &length, &err);

     if(ret == FALSE) {
          message_error("Fehler beim Oeffnen",
              g_strdup_printf("Konnte die Datei '%s' nicht oeffnen: %s", filename,
                  err->message)
          );
          return;
     }

     if(startsequence_running) {
         // Startsequenz (wenn Programm ohne Lochstreifendatei gestartet wird) laeuft noch
         // beenden, in dem startsequence_running auf FALSE gesetzt wird.
         startsequence_running = FALSE;
     }

     lochstreifen_set_data(lochstreifen, length, (byte_t *)data, -1, -1);
     message_statusbar("Die Datei wurde geoeffnet.");
     gtk_window_set_title(GTK_WINDOW (window),
         g_strdup_printf("%s - %s", basename(filename), WINDOW_TITLE)
    );

     // Neuzeichnen, usw.
     redraw_lochstreifen(TRUE);
}

gboolean open_lochstreifen_file_dialog(GtkWidget *widget, GtkWidget *parentWindow) {
    /**
     * Zeigt den Datei-Oeffnen-Dialog an (als Callback-Funktion fuer das Datei-Menue)
     *
     *
     **/
    GtkWidget *chooser;
    chooser = gtk_file_chooser_dialog_new(
        "(Binaer-)datei zur Darstellung als Lochstreifen auswaehlen",
        GTK_WINDOW(parentWindow),
        GTK_FILE_CHOOSER_ACTION_OPEN,
        GTK_STOCK_CANCEL, GTK_RESPONSE_CANCEL,
        GTK_STOCK_OPEN, GTK_RESPONSE_OK,
        NULL);
    if(gtk_dialog_run(GTK_DIALOG(chooser)) == GTK_RESPONSE_OK) {
        char *filename;
        FILE *file;

        filename = gtk_file_chooser_get_filename(GTK_FILE_CHOOSER (chooser));
        if(filename == NULL) {
            message_error("Keine Datei geoeffnet.", NULL);
            gtk_widget_destroy(GTK_WIDGET(chooser));
            return FALSE;
        }

        /*file = fopen(filename, "r");
        if(file == NULL) {
            message_error("Fehler beim Oeffnen",
                g_strdup_printf("Konnte die Datei '%s' nicht oeffnen: %s", filename,
                    g_strerror (errno))
            );
            return;
        }
        open_lochstreifen_file(file);*/
        open_lochstreifen_file(filename);
        g_free(filename);
    }
    gtk_widget_destroy(chooser);
    return FALSE;
}

void message_error(gchar *heading, gchar *text) {
    GtkWidget *dialog;
    dialog = gtk_message_dialog_new(GTK_WINDOW(window),
        GTK_DIALOG_MODAL,
        GTK_MESSAGE_ERROR,
        GTK_BUTTONS_OK,
        heading);
    if(text != NULL) gtk_message_dialog_format_secondary_text(GTK_MESSAGE_DIALOG(dialog), text);
    gtk_dialog_run(GTK_DIALOG(dialog));
    gtk_widget_destroy(GTK_WIDGET(dialog));
}


gboolean startup_sequence_loop(gpointer state) { //real_data_length) {
    // siehe startup_sequence
    //int state = *((int *)state_pointer);

    // Schleife wird deaktiviert, in dem startsequence_running auf false gesetzt wird.
    if(!startsequence_running)
        return FALSE;

    if(*((int*)state) <= startsequence_length) {
        lochstreifen->data++; // Zeiger zeigt aufs naechse Byte
        *((int*)state) += 1;
    } else {
        lochstreifen->data -= startsequence_length;
        *((int*)state) = 0;
    }


    // einfach erst mal die Lochstreifendaten um eins verschieben.
    /*if(lochstreifen->data_length < length) {
        lochstreifen->data_length++;
    }*/
    /*
    byte_t buf;
    int x;
    for(x = 1; x < lochstreifen->data_length;x++) {
        buf = lochstreifen->data[x-1];
        lochstreifen->data[x-1] = lochstreifen->data[x];
        lochstreifen->data[x] = buf;
    }*/

    // neuzeichnen.
    redraw_lochstreifen(TRUE);
    printf("Neugezeichnet.\n");

    return TRUE;
    // true zurueckgeben, wenns weitergehen soll.
}

gboolean startup_sequence(GtkWidget *widget, gpointer datas) {
    /**
     * Die kleine "Startup-Sequenz" wird gezeigt, wenn es keine Daten anzuzeigen gibt.
     * Dabei scrollt ein Text (eigene Darstellung) als Lochstreifen durch die Gegend.
     *
     **/
    // Integer, um aktuelle Position zu speichern
    int *state = malloc(sizeof(int));
    *state = 0;
    // Array mit gerade 2x Datenarray hintereinander erstellen.
    byte_t *bufdata = malloc(startsequence_length*2); // Buffer mit doppelter Laenge erstellen
    memcpy(bufdata, startsequence_data, startsequence_length); // Bufarray fuellen
    memcpy(bufdata+startsequence_length, startsequence_data, startsequence_length); // und ein zweites mal
    /*int x;
    byte_t *data = malloc(sizeof(byte_t)*256);
    for(x=0;x<256;x++) {
        data[x] = (byte_t) (255-x);
    }
    */

    /*int length_ = 256;
    for(x=0; x<length_; x++) {
        printf("%i von %i: 0x%x (=%c)\n", x, length_, data[x], data[x]);
    }*/

    lochstreifen_set_data(lochstreifen, startsequence_length, bufdata, 0, 0); // schon mal den ersten Wert.
    //*length = 256;
    g_timeout_add(1000, startup_sequence_loop, state); //length); // zur Main Loop hinzufuegen.
    return FALSE;
}

int main(int argc, char *argv[]) {
    GtkWidget *mainbox;
    GtkWidget *menubar, *menuitem, *menu;
    GtkSizeGroup *color_sizes; // Damit Farbmenue einheitlich aussieht
    GtkWidget *scroll;
    GtkWidget *statusbar;
    lochstreifen = lochstreifen_new(); // muss hier bereit zugewiesen werden weil benutzt

    gtk_init (&argc, &argv);

    window = gtk_window_new(GTK_WINDOW_TOPLEVEL);
    gtk_window_set_default_size (GTK_WINDOW (window), 600, 600);
    gtk_window_set_title(GTK_WINDOW (window), WINDOW_TITLE);
    g_signal_connect_swapped(G_OBJECT(window), "destroy", G_CALLBACK (gtk_main_quit), NULL);

    mainbox = gtk_vbox_new(FALSE, 0);
    gtk_container_add(GTK_CONTAINER (window), mainbox);
    gtk_widget_show(mainbox);

    statusbar = gtk_statusbar_new(); // erst nach Inhaltswidget hinzufuegen
    lochstreifen_statusbar = statusbar; // quick & dirty global machen...

    /* Men */
    menubar = gtk_menu_bar_new();
    gtk_box_pack_start (GTK_BOX (mainbox), menubar, FALSE, TRUE, 0);
    gtk_widget_show(menubar);

    // Erstes Menue: Datei
    menu = gtk_menu_new();
    // Oeffnen
    g_signal_connect(G_OBJECT(fast_stock_menuitem(menu, GTK_STOCK_OPEN)),
        "activate", G_CALLBACK(open_lochstreifen_file_dialog), window);
    fast_menu_seperator(menu);
    gtk_paper_tape_menu_export(menu);
    fast_menu_seperator(menu);
    g_signal_connect(G_OBJECT(fast_stock_menuitem(menu, GTK_STOCK_QUIT)),
        "activate", G_CALLBACK(gtk_main_quit), NULL);

    menuitem = gtk_menu_item_new_with_mnemonic("_Datei");
    gtk_menu_item_set_submenu(GTK_MENU_ITEM (menuitem), menu);
    gtk_menu_shell_append(GTK_MENU_SHELL (menubar), menuitem);
    gtk_widget_show (menuitem);

    // Zweites Menue: Ansicht
    menu = gtk_menu_new();
    gtk_paper_tape_menu_view(menu);

    menuitem = gtk_menu_item_new_with_mnemonic("_Ansicht");
    gtk_menu_item_set_submenu(GTK_MENU_ITEM (menuitem), menu);
    gtk_menu_shell_append(GTK_MENU_SHELL (menubar), menuitem);
    gtk_widget_show (menuitem);

    // Drittes Menue: Farben
    menu = gtk_menu_new();
    gtk_paper_tape_menu_colors(menu);

    menuitem = gtk_menu_item_new_with_mnemonic("_Farben");
    gtk_menu_item_set_submenu(GTK_MENU_ITEM (menuitem), menu);
    gtk_menu_shell_append(GTK_MENU_SHELL (menubar), menuitem);
    gtk_widget_show (menuitem);



    // Daten einlesen.
    {
        char *filename = NULL;
        gboolean read_from_stdin = FALSE;
        gboolean no_splash = FALSE;
        GError *error = NULL;
        GOptionContext *context;
        byte_t show_mystart = 0; // eigene Startprozedur anzeigen?
        GOptionEntry option_entries[] = {
           { "filename", 'f', 0, G_OPTION_ARG_FILENAME, &filename, "Datei zu oeffnen", NULL },
           { "stdin", 's', 0, G_OPTION_ARG_NONE, &read_from_stdin, "Von Standardeingabe lesen", NULL },
           { "no-splash", 'n', 0, G_OPTION_ARG_NONE, &no_splash,   "Keine Startanimation anzeigen, wenn ohne Datei gestartet", NULL },
           { NULL }
        };

        context = g_option_context_new(" - Lochstreifenvisualisierung");
        g_option_context_add_main_entries(context, option_entries, NULL);
        g_option_context_add_group(context, gtk_get_option_group(TRUE));
        g_option_context_parse(context, &argc, &argv, &error);

        if(read_from_stdin) {
            printf("%s: Lese Daten von Standardeingabe, erst nach EOF wird Fenster geoeffnet.\n",argv[0]);
            byte_t *data;
            int length = file_get_contents(stdin, &data);
            lochstreifen_set_data(lochstreifen, length, data, 6, 6);
        } else if(filename != NULL) {
             // eine Datei einlesen
             printf("Von Datei %s lesen\n", filename);
             if(!open_lochstreifen_file(argv[1]))
                 show_mystart = 1;
        } else if(!no_splash) {
            // Splash (=Start)-Sequenz anzeigen
            show_mystart = 1;
        }

        // Daten wurde eingelesen -- oder auch nicht:
        if(show_mystart != 0) {
            g_signal_connect_after(G_OBJECT(window), "show", G_CALLBACK(startup_sequence), NULL);
            startsequence_running = show_mystart;
        } else startsequence_running = FALSE;
    }

    papertape = gtk_paper_tape_new();
    gtk_box_pack_start(GTK_BOX(mainbox), papertape->scroll, TRUE, TRUE, 0);

    /// signal fr Statusbar verbinden zu lokaler Funktion!

    // Statusbar hinzufuegen.
    // Ich haette ja gerne eine maechtigere Statusbar, aber das geht wohl
    // leider nicht. Ein paar Versuche, das Label dort wenigstens Pango-Formatierungs-Faehig
    // zu machen:
    /*gtk_container_forall(GTK_CONTAINER(statusbar), G_CALLBACK(fast_nicer_statusbar), NULL);
    gboolean fast_nicer_statusbar(GtkWidget *statusbar_child, gpointer egal) {
        GList *children;
        gpointer child;
        children = gtk_container_get_children(GTK_CONTAINER(statusbar));
        
        gtk_container_forall                (GtkContainer *container,
                                                         GtkCallback callback,
                                                         gpointer callback_data);
        
        //printf("Checke Kindelement von statusbar (%d)\n", g_list_length(children));
        //while((child = g_list_next(children)) != NULL) {
            if(GTK_IS_LABEL(statusbar_child)) {
                printf("LABEL GEFUNDEN!\n");
                gtk_label_set_use_markup(GTK_LABEL(statusbar_child), TRUE); // :-)
            }
            printf("Fand ein %s-Objekt.\n", G_OBJECT_TYPE_NAME(statusbar_child));
        //}
        //printf("Fertig.\n");
    }    */
    gtk_box_pack_start(GTK_BOX(mainbox), statusbar, FALSE, TRUE, 0);
    gtk_widget_show(statusbar);

    // TEST DEBUG:
    //gtk_widget_set_redraw_on_allocate(lochstreifen_widget, TRUE);
    //gtk_widget_set_double_buffered(lochstreifen_widget, FALSE);

    gtk_widget_show(window);
    gtk_main();
    return 0;
}
