/**
 * Lochstreifen mit Cairo zeichnen
 *
 * Diese Datei stellt ein Objekt (LOCHSTREIFEN) zur Verfuegung,
 * welches mit Cairo Lochstreifen auf vielfaeltige Weise rendern
 * kann. Ueber die zahlreichen Funktionen koennen die Variablen
 * eingestellt werden.
 *
 * Diese Lochstreifenroutinen sind bewusst nur mit Cairo programmiert
 * worden (kompilieren mit `pkg-config --cflags --libs cairo`), sie
 * benoetigen kein installiertes GTK oder Glib.
 *
 * Die Variationsmoeglichkeiten fuer die gezeichneten Lochstreifen
 * sind sehr vielfaeltig und koennen einzeln voellig frei miteinander
 * kombiniert werden. So kann man saemtliche geometrischen Abstaende
 * aendern (Abstand der Lochreihen, Lochspalten, Fuehrungloecher,
 * Durchmesser der unterschiedlichen Loecher, Abstaende in und um
 * den Lochstreifen, usw.) oder verhaeltnissmaessig angeben.
 * Ausserdem laesst sich der Lochstreifen sehr leicht in 45-Schritten
 * drehen oder um beide Achsen spiegeln. Saemtliche Farben sind frei
 * einstellbar, einzelne Objekte koennen auf Wunsch auch nicht
 * gezeichnet werden.
 *
 * Anzumerken sei auch, dass diese Routinen nur die Zeichenarbeit
 * uebernehmen -- das Cairoobjekt muss das Programm stellen. Damit sind
 * voellige Freiheiten zur Wahl der Cairo-Surface gegeben (SVG, PNG,
 * PDF, ... -- alles moeglich, Cairo eben). Da man bei der Erstellung
 * einer Cairosurface ueblicherweise Groessenangaben braucht, kriegt
 * man diese aus den aktuellen Parametern berechnet mit
 * lochstreifen_get_width und lochstreifen_get_height.
 *
 * Fuer die Bedeutung der einzelnen Felder der LOCHSTREIFEN-Struktur
 * siehe Header-File, fuer genauere Erklaerungen zu den Funktionen
 * siehe direkt bei deren Implementierungen (diese Datei). 
 *
 * sven, im November 2007.
 *
 **/
#include <math.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "lochstreifen.h"

#define DEBUG 1

int file_get_contents(FILE *stream, byte_t **content) {
    /**
    * Kleine Behelfsfunktion, die den stream (z.B. stdin, eine Datei) als Bytes
    * komplett einliest.
    * Erwartet: Stream, Zeiger auf Zeiger (fr Array)
    * Gibt zurck:  Return-value:      Laenge von content, rechnend ab 1
    *               Call-by-reference: Content wird alloziiert als Array mit Inhalt
    *
    * Erzeugt seltsamerweise manchmal komische Fehler,
    * g_file_get_contents von Glib funktioniert zuverlaessiger.
    **/
    // das war meine Version, bevor ich mir alles von g_file_get_contents
    // abgeschaut habe. Einen Fehler hab ich zumindest gerade entdeckt:
    // realloc gibt ggf. einen neuen Pointer zurck. Das wrde erklren, warum auch
    // das ganze *mal* ging und *mal* nicht. Offensichtlich kopiert realloc dann auch
    // die Daten, wenn es entscheidet, sie woanders auf dem heap abzusieldeln...
    /*
    // um wie viel Zeichen das dynamische Array jeweils erhoeht werden soll
    // ein Kompromiss zwischen Platzverschwendung (zuviel allokatierter aber ungenutzter
    // Speicher) und Perfomance beim Einlesen (zu oft das Array erweitern)
    #define FILE_GET_CONTENT_STEPSIZE 20

    int length;
    int extended = 1; // wie oft bereits erweitert
    *content = (byte_t*) malloc(sizeof(byte_t) * FILE_GET_CONTENT_STEPSIZE);
    if(*content == NULL) return 0; // schlecht.

    for(length=0;;) {
        (*content)[length] = (byte_t)fgetc(stream);
        if(feof(stream) != 0) break; // EOF vor dem Einlesen!
        printf("%03d. Zeichen eingelesen: 0x%02X (=%c)\n", length+1, (*content)[length], (*content)[length]);
        if(extended*FILE_GET_CONTENT_STEPSIZE == ++length) {// Speicherende erreicht?
             printf("=> Bereits %d mal den %d-Puffer verwendet, erweitere auf %d...\n", extended, FILE_GET_CONTENT_STEPSIZE, (extended+1)*FILE_GET_CONTENT_STEPSIZE);
             if(realloc(*content, sizeof(byte_t)*(++extended)*FILE_GET_CONTENT_STEPSIZE) == NULL) // Speicher vergroessern
                 return length; // Fehler - das eingelesene zumindest zurueckgeben.
        }
    }

    return length; // wir wollen ab 1 zaehlen, das wurde indirekt durch den letzten unnoetigen ++inkrement gemacht
    }
    int file_get_contents(FILE *stream, byte_t **content) {
    */
    /**
     * Neugeschrieben nach etwas Inspiration von der glib am 05.04.2008.
     * Funktion get_contents_stdio, gfileutils.c im Sourcecode von glib-2.14.3.
     * Router natrlich aus (03:11!), aber da sieht man mal wieder den Vorteil von Gentoo:
     * Gleich alle Sourcecodes auf der Platte =)
     *
     *
     *
     **/
     // *stream = *f
     // *content = **contents
     // gsize *length

     char buf[4096];
     size_t bytes; // gerade eben eingelesene bytes
     char *str = NULL;
     size_t total_bytes = 0;     // alle bis jetzt eingelesenen bytes
     size_t total_allocated = 0;
     char *tmp; // fuers realloc

     while(!feof(stream)) {
         bytes = fread(buf, 1, sizeof(buf), stream);

         while( (total_bytes + bytes) > total_allocated) {
             if(str)
                 total_allocated *= 2;
             else
                 total_allocated = bytes > sizeof(buf) ? bytes : sizeof(buf);

             tmp = realloc(str, total_allocated);

             if(tmp == NULL) {
                 fprintf(stderr, "*** file_get_contents ERROR*** Could not reallocate\n");
                 //return length; // Fehler - das eingelesene zumindest zurueckgeben.
                 // nee, gar nichts zurckgeben. Das geht so nicht.
                 return 0;
             }

             str = tmp;
        } // while innen

        memcpy(str + total_bytes, buf, bytes);
        total_bytes += bytes;
     } // while

     if(total_allocated == 0)
         str = malloc(1); // ein leerer String halt.

     //str[total_bytes] = '\0'; // wenns ein string wre.

     *content = str;

     return total_bytes;
}

void lochstreifen_set_color(cairo_pattern_t *color, byte_t red, byte_t green, byte_t blue, byte_t alpha) {
    /** Kleine Befehlsfunktion, die ein Cairopattern erzeugt. Farbwerte zwischen 0 und 255.
        alpha = 255 ist unnoetig => einfach das Patternobjekt auf NULL setzen und die Zeichenfunktion
        wird es nicht zeichnen. */
    color = cairo_pattern_create_rgba(red/255, green/255, blue/255, alpha/255);
}


LOCHSTREIFEN *lochstreifen_new() {
    /** Neues Lochstreifenobjekt erzeugen */
    LOCHSTREIFEN *l = malloc(sizeof(LOCHSTREIFEN));
    // Defaultwerte setzen:
    l->d = 10;
    l->data_length = 0;
    l->data = NULL;
    l->empty_start = 0;
    l->empty_end = 0;
    l->debug = 0;
    l->hintergrund = NULL; // nicht zeichnen
    l->streifenbg = cairo_pattern_create_rgb(0.7,0.7,0.7);
    l->punched = cairo_pattern_create_rgb(0,0,0);
    l->notpunched = cairo_pattern_create_rgb(0.827, 0.827, 0.827); //220/255, 220/255, 220/255);
    l->fuehrung = cairo_pattern_create_rgb(143/255, 175/255, 255/255);
    //cairo_matrix_init_identity(&(l->transformation));
        //, 1, 0, 0, 1, 0, 0); // Standardmatrix einrichten (keine Drehung)
    l->drehung = 0;
    l->spiegelung_hor = 0;
    l->spiegelung_ver = 0;
    l->highlight_byte = 0;
    l->highlight_color = NULL;
    lochstreifen_flush_only_start_area(l);
    return l;
}

void lochstreifen_set_data(LOCHSTREIFEN *l, int data_length, byte_t *data, int empty_start, int empty_end) {
    /** Zu druckende Daten eingeben, Anzahl gewuenschter Leerbytes vorne/hinten eingeben,
        ist empty_start/empty_end < 0, wird das entsprechende nicht gesetzt */
    l->data_length = data_length;
    if(l->data != NULL) // unbenutzten Speicher freigeben!
        free(l->data);
    l->data = data;
    if(empty_start >= 0) l->empty_start = empty_start;
    if(empty_end >= 0) l->empty_end = empty_end;
}

void lochstreifen_set_highlight(LOCHSTREIFEN* l, int byte_number) {
    l->highlight_byte = byte_number;
    if(l->highlight_color == NULL)
        l->highlight_color = cairo_pattern_create_rgb(242/255, 255/255, 0);
}

int lochstreifen_set_only_start_area(LOCHSTREIFEN *l, int x, int y, int width, int height) {
    /** Clipping-Rechteck setzen. Damit wird beim naechsten Zeichnen nur in diesem
        Bereich gezeichnet. Es handelt sich jedoch um kein exiplites ordentliches
        cairo_clipping, sondern eine grobe Naehrung zwecks Perfomance.
        Die Koordinaten beziehen sich auf die entgueltige Form (inkl. Drehung, ...)
    */
    l->only_start_x = x;
    l->only_start_y = y;
    l->only_width = width;
    l->only_height = height;
}

int lochstreifen_flush_only_start_area(LOCHSTREIFEN *l) {
    /** Clipping-Rechteck loeschen -- beim naechsten Zeichnen alles zeichnen */
    l->only_start_x = 0;
    l->only_start_y = 0;
    l->only_width = 0;
    l->only_height = 0;
}

void lochstreifen_set_d(LOCHSTREIFEN *l, int d) {
    /** Durchmesser eines Loches uebergeben (allgemeine Referenzeinheit fuer Dimensionen) */
    l->d = d;
    l->bspace = d*0.7/1.9;  // Abstand der Bytes voneinander (Reihen)
    l->hspace = d/2;  // Abstand der Loecher in der Bytereihe
    l->padding = d; // Abstand Lcher <=> Lochstreifenrand
    l->df = d/2; // Durchmesser Lochfuehrungsloch
    l->fspace = 1.1*(d+ l->hspace); // Breite der Lochfuehrungs"nut"
    l->margin = d*2; // Abstand um Lochstreifen herum (an allen Seiten)
    l->abriss = (8*d+6* l->hspace + l->fspace)/2; // Lnge der Abrisse an Start/Ende (halbe Hhe vom Lochstreifen)
}

int lochstreifen_get_orientation(LOCHSTREIFEN *l) {
    /** Gibt die Orientierung des Lochstreifens zurueck; 1 = horizontal, 0 = vertikal */
    return (l->drehung % 2 == 0) ? 1 : 0;
}

int lochstreifen_get_width(LOCHSTREIFEN *l) {
    /** Berechnet aktuelle Breite aus Lochstreifenkonstanten */
    // interne Ersetzungen...
    #define LOCHSTREIFEN_WIDTH  2*l->abriss + 2*l->margin + \
        (l->data_length + l->empty_start + l->empty_end) * (l->d + l->bspace) - l->bspace // ein bspace zuviel am Ende
    #define LOCHSTREIFEN_HEIGHT 8*l->d + 6*l->hspace + l->fspace + l->margin*2 + l->padding*2 \
                                     //  ^ weil letztes hspace + hspace ber fspace unntig
    return (l->drehung % 2 == 0) ? LOCHSTREIFEN_WIDTH : LOCHSTREIFEN_HEIGHT;
}

int lochstreifen_get_height(LOCHSTREIFEN *l) {
    /** Berechnet aktuelle Hoehe aus Lochstreifenkonstanten */
    return (l->drehung % 2 == 1) ? LOCHSTREIFEN_WIDTH : LOCHSTREIFEN_HEIGHT;
}

void lochstreifen_set_d_by_width(LOCHSTREIFEN *l, int width) {
    /**
     * Umkehrfunktion zu set_d (dort wird die Breite/Hoehe durch d berechnet),
     * hier wird d durch die vorgegebene Breite berechnet. Die Hoehe ergibt
     * sich dann natuerlich auch automatisch.
     * Evventuelle Drehungen werden miteinberechnet, d.h. wenn der Lochstreifen
     * z.B. nach oben/unten zeigt, bezieht sich width auf die "echte" Hoehe.
     **/
    // Gleichung zum Berechnen von LOCHSTREIFEN_WIDTH nach
    // d aufgeloest mit MuPad Pro 3.2 ( ;-) )
    #define LOCHSTREIFEN_INV_WIDTH  (double) (1 / (6187./380. + 26./19. * (double)(l->data_length + l->empty_start + l->empty_end)))
    #define LOCHSTREIFEN_INV_HEIGHT  (double) (1 * 20./373.)
    lochstreifen_set_d(l,
        (int)rint((double)width * ((l->drehung % 2 == 0) ? LOCHSTREIFEN_INV_WIDTH : LOCHSTREIFEN_INV_HEIGHT))
    );
    //printf("Width soll %d ==> %d | %d\n", width, lochstreifen_get_width(l), lochstreifen_get_height(l));
}

void lochstreifen_set_d_by_height(LOCHSTREIFEN *l, int height) {
    /**
     * Siehe set_d_by_width, nur wird hier d abhaengig von der gewuenschten Hoehe
     * berechnet.
     **/
    lochstreifen_set_d(l,
        (int)rint((double) height * LOCHSTREIFEN_INV_HEIGHT) //((l->drehung % 2 == 0) ? LOCHSTREIFEN_INV_WIDTH : LOCHSTREIFEN_INV_HEIGHT))
    );
    //printf("Height soll %03d ==> d=%02d | %f, w/h = %03d | %03d\n", height, l->d, 
    //    (double)height*LOCHSTREIFEN_INV_HEIGHT
    //    , lochstreifen_get_width(l), lochstreifen_get_height(l));
}

void lochstreifen_set_direction(LOCHSTREIFEN *l, int drehung, int spiegelung_hor, int spiegelung_ver) {
    /**
     * Transformationen der Lochstreifenausrichtung
     * drehung: 0 = Kopf nach links
     *          1 = oben
     *          2 = rechts
     *          3 = unten
     *         -1 = nicht aendern
     *          4 = um 90 im Uhrzeigersinn drehen
     *          5 = um 90 gegen Uhrzeigersinn drehen
     * Spiegelung horizontal/vertikal: 0 = nein
     *                                 1 = ja
     *                                 2 = vertauschen
     *                                -1 = nicht aendern
     **/
     if(drehung >= 0) {
         // Das "Ausprobieren" zwecks "Autodrehung" mit Aufruf wie
         // lochstreifen_set_direction(l, l->drehung++, -1, -1);
         //if(drehung > 4) drehung -= 4; // bei zu grossen Werten mal ausprobieren
         //if(drehung < 0) drehung += 4; // bei zu kleinen Werten mal ausprobieren
         //if(drehung % 4 == 0) drehung = 0; // quasi keine Drehung
         if(drehung >= 4) {
             if(drehung == 4) l->drehung++;
             else if(drehung == 5) l->drehung--;
             if(l->drehung >= 4) l->drehung -= 4; // bei zu grossen Werten
             if(l->drehung < 0) l->drehung += 4; // bei zu kleinen Werten
         } else {
             //if(drehung > 0 && drehung < 4) // gute Werte
             // normale Drehung mit festem Zielwert
             l->drehung = drehung;
         }
     }
     if(spiegelung_hor > -1) {
         if(spiegelung_hor == 2) l->spiegelung_hor = l->spiegelung_hor == 0 ? 1 : 0; // negotieren
         else l->spiegelung_hor = spiegelung_hor; // normal
     }
     if(spiegelung_ver > -1) {
         if(spiegelung_ver == 2) l->spiegelung_ver = l->spiegelung_ver == 0 ? 1 : 0; // negotieren
         else l->spiegelung_ver = spiegelung_ver; // normal
     }
}

/*
void lochstreifen_rotate(LOCHSTREIFEN *l) {
    **
     * Wird den Lochstreifen beim naechsten Zeichnen um 45 mit dem
     * Uhrzeigersinn rotiert anzeigen. Nur mit diesen Funktionen wird
     * die Hoehen/Breitenberechnung im Vorhinein noch funktionieren!
     **
     cairo_matrix_rotate(&(l->transformation), M_PI);
     cairo_matrix_translate((&l->transformation), lochstreifen_get_width(l), 0),
     l->drehung++; // ...
}

void lochstreifen_spiegel(LOCHSTREIFEN *l, byte_t horizontal) {
    **
     * Spiegelt den Lochstreifen, ist horizontal = 1 dann horizontal,
     * sonst vertikal (oder so aehnlich)
     **
     if(horizontal != 0) { // vertikal
         cairo_matrix_scale(&(l->transformation), -1, 1);
         cairo_matrix_translate((&l->transformation), 0, lochstreifen_get_height(l));
     } else {
         cairo_matrix_scale(&(l->transformation), 1, -1);
         cairo_matrix_translate((&l->transformation), lochstreifen_get_width(l), 0);
     }
}
*/

int lochstreifen_byte_by_coordinate(LOCHSTREIFEN *l, int x, int y) {
     /**
      * Anhand der Koordinaten (x|y) wird der Index des Bytes (fuers Datenarray, ab 1 zaehlend),
      * welches damit "erfasst" wird, zurueckgegeben.
      * Befindet sich (x|y) nicht im Bereich eines Bytes, dann wird je nach Fall zurueckgegeben:
      *    0 = ausserhalb Lochstreifens
      *   -1 = in Emptystart/Emptyend-Bytes (die nicht in der Datei existieren)
      *        oder in Abrissbereichen
      **/
    int byte;
    // BETRACHTUNG HORIZONTALE AUSRICHTUNG

    if(y <= l->margin || y >= LOCHSTREIFEN_HEIGHT - l->margin) return 0; // ueber/unter Lochstreifen
    x -= l->margin; // erst mal Rand abziehen
    if(x <= 0) return 0; // links neben dem Lochstreifen.
    x -= l->abriss + (l->empty_start)*(l->d + l->bspace); // Emptystart+Abrissbereich
    if(x <= 0) return -1; // in diesem Bereich gewesen
    // jetzt sind wir im Bytebereich
    byte = (int) (x / (l->d + l->bspace)) + 1; // Abrunden: Auf welchem Byte man ist + ab 1 zaehlen.
    if(byte > l->data_length + l->empty_start) return 0; // komplett ausserhalb Lochstreifens
    else if(byte > l->data_length) return -1; // im Emptyendbereich
    else return byte; // im gueltigen Bereich -- Byte gefunden!
}

int lochstreifen_coordinate_by_byte(LOCHSTREIFEN *l, int byte) {
    /**
     * Anhang eines Bytes (fuers Datenarray, ab 0 zaehlend) wird -- abhaengig nach aktueller
     * Ausrichtung des Lochstreifens -- die Koordinate des Bytes in der Dimension der
     * Ausrichtung des Lochstreifens zurueckgegeben..
     * Liegt der Lochstreifen horizontal, so wird die x-Koordinate zurueckgegeben. Liegt er
     * vertikal, dann die y-Koordinate.
     *
     **/
     int xy;
     //if(lochstreifen_get_orientation(l)) { // horizontal = 1 | vertical = 0
        xy = l->margin + l->abriss + (l->d + l->bspace) * (byte + l->empty_start);
     //}

     return xy;
     // TODO! Richtige implementierung!
}

void lochstreifen_draw(LOCHSTREIFEN *l, cairo_t *cr) {
    /**
     * Den Lochstreifen auf das uebergebene Cairoobjekt zeichnen.
     * Die Funktion zeichnet nach den aktuellen Lochstreifenobjekt-Angaben
     * (Groesse/Breite) -- wenn die cairo_image_surface nicht diesen Dimensionen
     * entspricht, ist das alt schlecht.
     *
     **/
    int byte,loch,x,y; // temp fuer Schleifen
    int height = lochstreifen_get_height(l);
    int width = lochstreifen_get_width(l);
    // 4 Eckpunkte des Zeichenbereichs bestimmen...
    int x1,x2,y1,y2;
    x1 = l->only_start_x; y1 = l->only_start_y;
    x2 = x1 + width; y2 = y1 + height;

    /*int x1,y1,x2,y2,x3,y3,x4,y4;
    x1 = l->only_start_x; y1 = l->only_start_y;
    x2 = x1+width; y2 = y1;
    x3 = x2; y3 = y2+height;
    x4 = x1; y4 = y3;*/

    // Verschiebungsmatrix erst hier berechnen, weil Verschiebungen vorher unklar
    // wunderbare affine Abbildungen ;-)
    cairo_matrix_t matrix, temp; // temp: Temporaere Matrix zur Matrizenmulitiplikation
    cairo_matrix_init_identity(&matrix);
    if(l->drehung > 0) {
        if(l->drehung == 1) {
            cairo_matrix_init(&temp, 0, -1, 1, 0, 0, height);
            x = width; width = height; height = x;                // Breiten und Hoehen
            x = l->only_width; l->only_width = l->only_height; l->only_height = x; // gerade vertauschen
        } else if(l->drehung == 2) {
            cairo_matrix_init(&temp, -1, 0, 0, -1, width, height);
        } else if(l->drehung == 3) {
            cairo_matrix_init(&temp, 0, 1, -1, 0, width, 0);
            x = width; width = height; height = x;                // Breiten und Hoehen
            x = l->only_width; l->only_width = l->only_height; l->only_height = x; // gerade vertauschen
        }
        cairo_matrix_multiply(&matrix, &matrix, &temp);
    }

    if(l->spiegelung_hor != 0) {
         cairo_matrix_init(&temp, -1, 0, 0, 1,
             l->drehung % 2 == 0 ? width : height, // Default bei Nicht-Drehung: width
             0);
         cairo_matrix_multiply(&matrix, &matrix, &temp);
    }
    if(l->spiegelung_ver != 0) {
         cairo_matrix_init(&temp, 1, 0, 0, -1, 0,
             l->drehung % 2 == 0 ? height: width); // Default bei Nicht-Drehung: height
         cairo_matrix_multiply(&matrix, &matrix, &temp);
    }

    //if(l->drehung > 0) {
        /*cairo_matrix_rotate(&matrix, l->drehung/2 *M_PI);
        if(l->drehung == 1) { // 90, im III. Sektor
            //x = width; width = height; height = x;
            //cairo_matrix_translate(&matrix, 0, 0);
        } else if(l->drehung == 2) { // 180, im IV Sektor
            cairo_matrix_translate(&matrix, width, height);
        } else if(l->drehung == 3) { // 270, im I. Sektor
            //x = width; width = height; height = x;
            cairo_matrix_translate(&matrix, width, height);
        }*/
        // 4. Sektor kommt nicht vor
    //}
    cairo_set_matrix(cr, &matrix);
    // -1, 0, 0, 1, width, 0 = spiegelung x-achse

    //cairo_device_to_user(cr, (double*)&(l->only_start_x), (double*)&(l->only_start_y));
    cairo_device_to_user(cr, (double*)&(x1), (double*)&(y1));
    cairo_device_to_user(cr, (double*)&(x2), (double*)&(y2));

    l->only_start_x = x2 < x1 ? x2 : x1;
    l->only_start_y = y2 < y1 ? y2 : y1;

    cairo_set_source(cr, l->fuehrung);
    cairo_rectangle(cr, l->only_start_x, l->only_start_y, l->only_width, l->only_height);
    //printf("Only: %d|%d, %d * %d\n", l->only_start_x, l->only_start_y, l->only_width, l->only_height);
    cairo_paint(cr);

    // Hintergrundfarbe zeichnen.
    if(0||l->hintergrund != NULL) {
        //l->hintergrund = cairo_pattern_create_rgb(1,1,1);
        cairo_set_source(cr, l->hintergrund);
        cairo_paint(cr);
    }

    // sache zeichnen
    cairo_set_source(cr, l->streifenbg);
    cairo_rectangle(cr, -10, -10, width+20, height+20);
    cairo_stroke(cr);
    if(0==1){
        double x,y, a,b;
        int xx,yy,aa,bb;
        x=0;y=0; a=width;b=height;
        cairo_matrix_transform_point(&matrix, &x, &y);
        cairo_matrix_transform_point(&matrix, &a, &b);
        xx=(int)x;yy=(int)y;aa=(int)a;bb=(int)b;
        printf("Punkt (0/0)=(%i|%i) und (%i|%i)=(%i/%i)\n", xx, yy, width, height, aa, bb);
    }

    // Lochstreifenhintergrund inklusive Abriss vorne/hinten malen
    if(l->streifenbg != NULL) {
        cairo_set_source(cr, l->streifenbg);

        // Abriss vorne
        cairo_move_to(cr, l->margin + l->abriss*0.8, l->margin); // oben rechts
        cairo_line_to(cr, l->margin, l->margin + l->padding+ 3*l->d + 2*l->hspace + l->fspace/2); // genau die fspace-Mitte
        cairo_line_to(cr, l->margin + l->abriss, height - l->margin); // unten ganz rechts
        cairo_line_to(cr, l->margin + l->abriss, l->margin); // nach oben ganz rechts...
        cairo_close_path(cr); // schliessen
        // normaler Bereich
        cairo_rectangle(cr, l->margin + l->abriss, l->margin,
            width - 2*l->margin - 2*l->abriss, height - 2*l->margin);
        /*printf("Zeichne Rechteck x|y = (%d|%d) width*height = %d * %d\n",
            l->margin + l->abriss, l->margin,
            width - 2*l->margin - 2*l->abriss, height - 2*l->margin
        );*/
        // Abriss hinten...
        cairo_move_to(cr, width - l->margin - l->abriss*0.2, l->margin); // oben mitte
        cairo_line_to(cr, width - l->margin - l->abriss,
            l->margin + l->padding + 3*l->d+ 2*l->hspace + l->fspace/2); // genau die fspace-Mitte
        cairo_line_to(cr, width - l->margin, height - l->margin); // unten ganz rechts
        cairo_line_to(cr, width - l->margin - l->abriss, height - l->margin); // unten ganz links
        cairo_line_to(cr, width - l->margin - l->abriss, l->margin); // nach oben zum schliessen
        cairo_close_path(cr);

        cairo_fill(cr);
    }


    // Lochstreifenlcher malen

    /*if(l->only_start_x > 0) {
        // nur in bestimmten Bereich zeichnen
        int start_byte = l->empty_start + lochstreifen_byte_by_coordinate(l, l->only_start_x, l->only_start_y); // absolut inkl. empty bytes

    } else byte = -l->empty_start; // leere Bytes vorne eingeschlossen
    }*/

    // Bytes durchgehen, leere Bytes vorne/hinten mit eingeschlossen
    for(byte = -l->empty_start, x=l->margin + l->abriss; byte < l->data_length + l->empty_end; x += l->d + l->bspace, byte++) {
        // Clipping: Um unnoetiges Zeichnen zu ersparen, erst ab X-Koordinate zeichnen
        if(x < l->only_start_x - (l->d + l->bspace)) continue; // - 1x Loch weniger, damit nicht ein gerade abgeschnittenes nichtgezeichnet wird
        // und nur bis zur rechten Begrenzung zeichnen -- dannach genuegt Abbruch.
        if(l->only_width != 0 && x > l->only_start_x + l->only_width) break;

        // ggf. Byte hervorheben
        if(byte == l->highlight_byte && l->highlight_color != NULL) {
            cairo_set_source(cr, l->highlight_color);
            cairo_rectangle(cr, x - l->bspace/2, l->margin, l->df*2 + l->bspace, height - 2*l->margin);
            cairo_fill(cr);
        }

        // Bits des entsprechenden Bytes durchgehen
        for(loch=0, y = l->margin + l->padding; loch < 8; y += l->d + l->hspace, loch++) {
            // Clipping: Erst ab Y-Koordinate zeichnen (-1xLoch-Korrektur gegen Abschnitt)
            if(y < l->only_start_y - (l->d - l->hspace)) continue;
            // bis zur unteren Begrenzung -- anschliessend Abbruch
            if(l->only_height != 0 && y > l->only_start_y + l->only_height) break;

            // Ggf. Lochstreifenfuehrung zeichnen
            if(loch==3 && l->fuehrung != NULL) {
                // Nur vor dem 3. Loch (wenn ueberhaupt)
                y -= l->hspace; // den Abstand vom vorherigen Loch wieder wegnehmen
                cairo_set_source(cr, l->fuehrung);
                cairo_arc(cr, x+l->d/2, y+l->fspace/2, l->df/2, 0., 2*M_PI);
                cairo_fill(cr);
                y += l->fspace;
            } else if(loch == 3) // 3. Loch, aber keine Fuehrung zeichnen => Platz freilassen!
                y += l->fspace - l->hspace;

            // Debugausgaben machen?
            if(l->debug != 0 && byte >= 0 && byte < l->data_length) {
                if(loch == 0) printf("byte %3i von %3i: 0x%2x = ", byte, l->data_length, l->data[byte]);
                fprintf(stderr, "%c", (((l->data[byte] >> loch) & 0x01) == 0x01) ? '#' : ' ');
                if(loch == 7) printf("\n");
            }

            if((byte < 0 || byte >= l->data_length) ||    // wenn ein Nullbyte gezeichnet wird
               (((l->data[byte] >> loch) & 0x01) != 0x01) // oder das Bit an der Stelle nicht gesetzt ist
               ) {                                        // ist es not-Punched
                if(l->notpunched != NULL) { // notpunched zeichnen? (muss subif sein)
                    cairo_set_source(cr, l->notpunched);
                    cairo_arc(cr, x+l->d/2, y+l->d/2, l->d/2, 0., 2*M_PI);
                    cairo_fill(cr);
                }
            } else if(l->punched != NULL) { // es ist Punched. Zeichnen?
                cairo_set_source(cr, l->punched);
                cairo_arc(cr, x+l->d/2, y+l->d/2, l->d/2, 0., 2*M_PI);
                cairo_fill(cr);
            }
            // Die Routine wegen dem Zeichnen?-Ding hervorgehoben.
            /*cairo_set_source(cr, 
                 (byte < 0 || byte >= data_length) ||      // wenn ein Nullbyte gezeichnet wird
                 (((data[byte] >> loch) & 0x01) != 0x01) ? // oder das Bit an der Stelle nicht gesetzt ist
                 notpunched : punched);                    // dann dies farbig kennzeichnen
            cairo_arc(cr, x+d/2, y+d/2, d/2, 0., 2*M_PI);
            cairo_fill(cr);*/

            // y+= war hier.
        } // for bits (loecher)
        // x += war hier.
    } // for bytes

    // Lochstreifenfuehrung im vorderen Abriss malen
    if(l->fuehrung != NULL) {
        cairo_set_source(cr, l->fuehrung);
        y = l->margin + l->padding + 3*l->d + 2*l->hspace + l->fspace/2; // fspace-Mitte
        x = l->margin + l->abriss - l->d - l->bspace; // Start-x
        for(; x > l->margin;) {
            cairo_arc(cr, x + l->d/2, y, l->df/2, 0., 2*M_PI);
            x -= l->d + l->bspace;
        }
        cairo_fill(cr);
    }

    // fertig mit dem Lochstreifenzeichnen.
} // function draw_lochstreifen
