/*
** Allgemeine Headerfile fuer das
** Lochstreifen-Visualisieren-Projekt
*/
#include <cairo.h> // cairo_t unten

#define byte_t unsigned char
#define LOCHSTREIFEN struct lochstreifen

// Die Lochstreifenstruktur, die alle relevanten Daten enthaelt
struct lochstreifen {
    /* Daten, die der Lochstreifen enthaelt: */
    int data_length; // Lnge des Datenarrays, rechnend ab 1 (wie argc)
    byte_t *data;    // Datenarray

    int empty_start; // wie viele 0bytes vorne
    int empty_end; // wie viele 0bytes hinten

    /* Konstanten zum Aussehen des Lochstreifens
       Referenzeinheit ist d, daraus wird alles berechnet,
       siehe lochstreifen_set_d
    */
    int d; // Durchmesser eines Loches
    int bspace;  // Abstand der Bytes voneinander (Reihen)
    int hspace;  // Abstand der Loecher in der Bytereihe
    int padding; // Abstand Lcher <=> Lochstreifenrand
    int df; // Durchmesser Lochfuehrungsloch
    int fspace; // Breite der Lochfuehrungs"nut"
    int margin; // Abstand um Lochstreifen herum (an allen Seiten)
    int abriss; // Lnge der Abrisse an Start/Ende (halbe Hhe vom Lochstreifen)

    /* Farben, in denen der Lochstreifen gehalten ist */
    cairo_pattern_t *hintergrund, *streifenbg, *punched, *notpunched, *fuehrung;
    // Farben um Lochstreifen, Lochstreifen, Loecher, Nicht-Loecher, Fuehrung
    // siehe zum Setzen: lochstreifen_set_color oder per Cairo-Funktion fuer
    // komplexere Loesungen. Bei NULL wird die Komponente nicht gezeichnet!

    /* Drehungen und co, nicht direkt veraendern, sonst stimmt die vorberechnete
       Hoehe/Breite nicht mehr mit den realen Werten ueberein. */
    //cairo_matrix_t transformation;
    // Das folgende koennte man natuerlich effizienter Speichern, aber egal.
    int drehung; // 0-3, jeweils 45-Schritte im Uhrzeigersinn
    int spiegelung_hor; // horizontale Spiegelung != 0 == ja
    int spiegelung_ver; // vertikale Spiegelung != 0 == ja

    /* Nur bestimmte rechteckige Bereiche zeichnen -- dies wird massgeblich vom
       GTK-Interface benutzt */
    int only_start_x; // Die angegebenen Punkte
    int only_start_y; // spannen ein Rechteck auf. In grober
    int only_width;   // Naeherung wird nur in diesem Bereich
    int only_height;  // gezeichnet.


    /* Feature: Ein bestimmtes Byte hervorheben */
    int highlight_byte; // Nummer des Bytes (von 0)
    cairo_pattern_t *highlight_color; // Farben => wenn NULL, kein Highlight

    /* Debugausgaben machen => wenn debug != 0 ist! */
    byte_t debug;
};

/* lochstreifen.c */
// allgemeine Zusatzfunktion, kein Bezug
int file_get_contents(FILE *stream, byte_t **content);

// Lochstreifenfunktionen, siehe Dokumentation in c-File
LOCHSTREIFEN *lochstreifen_new();
void lochstreifen_set_data(LOCHSTREIFEN *l, int data_length, byte_t *data, int empty_start, int empty_end);
void lochstreifen_set_d(LOCHSTREIFEN *l, int d);
void lochstreifen_set_d_by_width(LOCHSTREIFEN *l, int width);
void lochstreifen_set_d_by_height(LOCHSTREIFEN *l, int height);
int lochstreifen_get_width(LOCHSTREIFEN *l);
int lochstreifen_get_height(LOCHSTREIFEN *l);
int lochstreifen_set_draw_only_area(LOCHSTREIFEN *l, int x, int y, int width, int height);
int lochstreifen_flush_draw_only_area(LOCHSTREIFEN *l);
void lochstreifen_set_color(cairo_pattern_t *color, byte_t red, byte_t green, byte_t blue, byte_t alpha);
//void lochstreifen_rotate(LOCHSTREIFEN *l);
//void lochstreifen_spiegel(LOCHSTREIFEN *l, byte_t horizontal);
void lochstreifen_set_direction(LOCHSTREIFEN *l, int drehung, int spiegelung_hor, int spiegelung_ver);
void lochstreifen_draw(LOCHSTREIFEN *l, cairo_t *cr);
int lochstreifen_get_orientation(LOCHSTREIFEN *l);
int lochstreifen_byte_by_coordinate(LOCHSTREIFEN *l, int x, int y);
